# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, List, Optional, Set
import abc
import dataclasses
import datetime
import enum
import logging

from basana.backtesting import helpers, liquidity, value_map
from basana.core import bar, logs
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


logger = logging.getLogger(__name__)


@enum.unique
class OrderState(enum.Enum):
    OPEN = 100
    COMPLETED = 101
    CANCELED = 102


@dataclasses.dataclass
class OrderInfo:
    #: The order id.
    id: str
    #: True if the order is open, False otherwise.
    is_open: bool
    #: The operation.
    operation: OrderOperation
    #: The original amount.
    amount: Decimal
    #: The amount filled.
    amount_filled: Decimal
    #: The amount remaining.
    amount_remaining: Decimal
    #: The quote amount filled.
    quote_amount_filled: Decimal
    #: The fees charged.
    fees: Dict[str, Decimal]
    #: The limit price.
    limit_price: Optional[Decimal] = None
    #: The stop price.
    stop_price: Optional[Decimal] = None
    #: The ids of the associated loans.
    loan_ids: List[str] = dataclasses.field(default_factory=list)

    @property
    def fill_price(self) -> Optional[Decimal]:
        """The fill price."""
        fill_price = None
        if self.amount_filled:
            fill_price = self.quote_amount_filled / self.amount_filled
        return fill_price


@dataclasses.dataclass
class Fill:
    when: datetime.datetime
    balance_updates: Dict[str, Decimal]
    fees: Dict[str, Decimal]


# This is an internal abstraction to be used by the exchange.
class Order(metaclass=abc.ABCMeta):
    def __init__(
            self, id: str, operation: OrderOperation, pair: Pair, amount: Decimal, state: OrderState,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        assert amount > Decimal(0), f"Invalid amount {amount}"

        self._id = id
        self._operation = operation
        self._pair = pair
        self._amount = amount
        self._state = state
        self._balance_updates = value_map.ValueMap()
        self._fees = value_map.ValueMap()
        self._fills: List[Fill] = []
        self._auto_borrow = auto_borrow
        self._auto_repay = auto_repay
        self._loan_ids: Set[str] = set()

    @property
    def id(self) -> str:
        return self._id

    @property
    def pair(self) -> Pair:
        return self._pair

    @property
    def amount(self) -> Decimal:
        return self._amount

    @property
    def operation(self) -> OrderOperation:
        return self._operation

    @property
    def state(self) -> OrderState:
        return self._state

    @property
    def is_open(self) -> bool:
        return self._state == OrderState.OPEN

    @property
    def balance_updates(self) -> value_map.ValueMap:
        return self._balance_updates

    @property
    def fees(self) -> value_map.ValueMap:
        return self._fees

    @property
    def amount_filled(self) -> Decimal:
        return abs(self._balance_updates.get(self.pair.base_symbol, Decimal(0)))

    @property
    def amount_pending(self) -> Decimal:
        return self._amount - self.amount_filled

    @property
    def quote_amount_filled(self) -> Decimal:
        return abs(self._balance_updates.get(self.pair.quote_symbol, Decimal(0)))

    @property
    def fills(self) -> List[Fill]:
        return self._fills

    @property
    def auto_borrow(self) -> bool:
        return self._auto_borrow

    @property
    def auto_repay(self) -> bool:
        return self._auto_repay

    def cancel(self):
        assert self._state == OrderState.OPEN
        self._state = OrderState.CANCELED

    def add_fill(self, when: datetime.datetime, balance_updates: Dict[str, Decimal], fees: Dict[str, Decimal]):
        self._balance_updates += balance_updates
        self._fees += fees
        if self.amount_filled >= self.amount:
            self._state = OrderState.COMPLETED
        self._fills.append(Fill(when=when, balance_updates=balance_updates, fees=fees))

    def add_loan(self, loan_id: str):
        self._loan_ids.add(loan_id)

    def get_order_info(self) -> OrderInfo:
        return OrderInfo(
            id=self.id, is_open=self._state == OrderState.OPEN, operation=self.operation,
            amount=self.amount, amount_filled=self.amount_filled, amount_remaining=self.amount_pending,
            quote_amount_filled=self.quote_amount_filled,
            fees={symbol: -amount for symbol, amount in self._fees.items() if amount},
            loan_ids=[loan_id for loan_id in self._loan_ids]
        )

    @abc.abstractmethod
    def get_balance_updates(
            self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy
    ) -> Dict[str, Decimal]:
        """
        Returns the balance updates required to fill the order.

        :param bar: The bar that summarizes the trading activity.
        :param liquidity_strategy: The strategy used to model available liquidity.
        :returns: A dictionary that maps the symbol to the amount.

        .. note::

            * It can be either a complete or partial fill, based on the trading activity summarized by the bar and the
              available liquidity.
            * It should include both the base amount and the quote amount, with opposite signs depending on the
              operation.
        """
        raise NotImplementedError()

    def calculate_estimated_fill_price(self) -> Optional[Decimal]:
        """
        Optionally override to return an estimate for the fill price.
        This will be used to reserve the funds that will be required later for processing the order.
        """
        return None

    def not_filled(self):
        """Called every time the order was processed but no fill took place."""
        pass

    def get_debug_info(self) -> dict:
        return {
            "id": self.id,
            "pair": self.pair,
            "operation": self.operation,
            "amount": self.amount,
            "amount_filled": self.amount_filled,
        }


class MarketOrder(Order):
    def __init__(
            self, id: str, operation: OrderOperation, pair: Pair, amount: Decimal, state: OrderState,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        super().__init__(id, operation, pair, amount, state, auto_borrow=auto_borrow, auto_repay=auto_repay)

    def not_filled(self):
        # Fill or kill market orders.
        self.cancel()

    def get_balance_updates(self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy) -> Dict[str, Decimal]:
        # No partial fills for market orders.
        if self.amount_pending > liquidity_strategy.available_liquidity:
            logger.debug(logs.StructuredMessage("Not enough liquidity to fill order", order_id=self.id))
            return {}

        amount = self.amount_pending
        base_sign = helpers.get_base_sign_for_operation(self.operation)
        if self.operation == OrderOperation.BUY:
            price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_high=bar.high)
        else:
            assert self.operation == OrderOperation.SELL
            price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_low=bar.low)

        return {
            self.pair.base_symbol: amount * base_sign,
            self.pair.quote_symbol: price * amount * -base_sign
        }


class LimitOrder(Order):
    def __init__(
            self, id: str, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            state: OrderState, auto_borrow: bool = False, auto_repay: bool = False
    ):
        assert limit_price > Decimal(0), "Invalid limit_price {limit_price}"

        super().__init__(id, operation, pair, amount, state, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._limit_price = limit_price

    def get_balance_updates(self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy) -> Dict[str, Decimal]:
        amount = min(self.amount_pending, liquidity_strategy.available_liquidity)
        if not amount:
            return {}

        price = None
        base_sign = helpers.get_base_sign_for_operation(self.operation)

        if self.operation == OrderOperation.BUY:
            # Limit price was hit at bar open.
            if bar.open < self._limit_price:
                price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_high=self._limit_price)
            # The price went down to limit price or lower.
            elif bar.low <= self._limit_price:
                price = self._limit_price
        else:
            assert self.operation == OrderOperation.SELL
            # Limit price was hit at bar open.
            if bar.open > self._limit_price:
                price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_low=self._limit_price)
            # The price went up to limit price or higher.
            elif bar.high >= self._limit_price:
                price = self._limit_price

        ret = {}
        if amount and price:
            ret = {
                self.pair.base_symbol: amount * base_sign,
                self.pair.quote_symbol: price * amount * -base_sign
            }
        return ret

    def calculate_estimated_fill_price(self) -> Optional[Decimal]:
        # It will be the limit price or a better one.
        return self._limit_price

    def get_debug_info(self) -> dict:
        ret = super().get_debug_info()
        ret["limit_price"] = self._limit_price
        return ret

    def get_order_info(self) -> OrderInfo:
        ret = super().get_order_info()
        ret.limit_price = self._limit_price
        return ret


class StopOrder(Order):
    def __init__(
            self, id: str, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal,
            state: OrderState, auto_borrow: bool = False, auto_repay: bool = False
    ):
        assert stop_price > Decimal(0), "Invalid stop_price {stop_price}"

        super().__init__(id, operation, pair, amount, state, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._stop_price = stop_price

    def not_filled(self):
        # Fill or kill stop orders.
        self.cancel()

    def get_balance_updates(self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy) -> Dict[str, Decimal]:
        # No partial fills for stop orders.
        if self.amount_pending > liquidity_strategy.available_liquidity:
            logger.debug(logs.StructuredMessage("Not enough liquidity to fill order", order_id=self.id))
            return {}

        price = None
        amount = self.amount_pending
        base_sign = helpers.get_base_sign_for_operation(self.operation)
        if self.operation == OrderOperation.BUY:
            # Stop price was hit at bar open.
            if bar.open >= self._stop_price:
                price = bar.open
            # The price went up to stop price or higher.
            elif bar.high >= self._stop_price:
                price = self._stop_price

            if price:
                price = slipped_price(price, self.operation, amount, liquidity_strategy, cap_high=bar.high)
        else:
            assert self.operation == OrderOperation.SELL
            # Stop price was hit at bar open.
            if bar.open <= self._stop_price:
                price = bar.open
            # The price went down to stop price or lower.
            elif bar.low <= self._stop_price:
                price = self._stop_price

            if price:
                price = slipped_price(price, self.operation, amount, liquidity_strategy, cap_low=bar.low)

        ret = {}
        if price:
            ret = {
                self.pair.base_symbol: amount * base_sign,
                self.pair.quote_symbol: price * amount * -base_sign
            }
        return ret

    def calculate_estimated_fill_price(self) -> Optional[Decimal]:
        # It should be around the stop price, or at least we hope so.
        return self._stop_price

    def get_debug_info(self) -> dict:
        ret = super().get_debug_info()
        ret["stop_price"] = self._stop_price
        return ret

    def get_order_info(self) -> OrderInfo:
        ret = super().get_order_info()
        ret.stop_price = self._stop_price
        return ret


class StopLimitOrder(Order):
    def __init__(
            self, id: str, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal,
            limit_price: Decimal, state: OrderState, auto_borrow: bool = False, auto_repay: bool = False
    ):
        assert stop_price > Decimal(0), "Invalid stop_price {stop_price}"
        assert limit_price > Decimal(0), "Invalid limit_price {limit_price}"

        super().__init__(id, operation, pair, amount, state, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._stop_price = stop_price
        self._limit_price = limit_price
        self._stop_price_hit = False

    def get_balance_updates_before_stop_hit(
            self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy
    ) -> Dict[str, Decimal]:
        assert not self._stop_price_hit

        price = None
        base_sign = helpers.get_base_sign_for_operation(self.operation)
        amount = min(self.amount_pending, liquidity_strategy.available_liquidity)

        if self.operation == OrderOperation.BUY:
            # Stop price was hit at bar open.
            if bar.open >= self._stop_price:
                self._stop_price_hit = True
                # Limit price was also hit at open.
                if bar.open <= self._limit_price:
                    price = bar.open
                # Limit price was hit some time later within the bar.
                elif bar.low <= self._limit_price <= bar.high:
                    price = self._limit_price
            # The price went up to stop price or higher.
            elif bar.high >= self._stop_price:
                self._stop_price_hit = True
                # Limit price was hit some time later within the bar.
                if bar.low <= self._limit_price <= bar.high:
                    price = self._limit_price
            # Calculate slippage if necessary.
            if price is not None and price != self._limit_price:
                price = slipped_price(price, self.operation, amount, liquidity_strategy, cap_high=self._limit_price)
        else:
            assert self.operation == OrderOperation.SELL
            # Stop price was hit at bar open.
            if bar.open <= self._stop_price:
                self._stop_price_hit = True
                # Limit price was also hit at open.
                if bar.open >= self._limit_price:
                    price = bar.open
                # Limit price was hit some time later within the bar.
                elif bar.low <= self._limit_price <= bar.high:
                    price = self._limit_price
            # The price went down to stop price or lower.
            elif bar.low <= self._stop_price:
                self._stop_price_hit = True
                # Limit price was hit some time later within the bar.
                if bar.low <= self._limit_price <= bar.high:
                    price = self._limit_price
            # Calculate slippage if necessary.
            if price is not None and price != self._limit_price:
                price = slipped_price(price, self.operation, amount, liquidity_strategy, cap_low=self._limit_price)

        ret = {}
        if amount and price:
            ret = {
                self.pair.base_symbol: amount * base_sign,
                self.pair.quote_symbol: price * amount * -base_sign
            }

        return ret

    def get_balance_updates_after_stop_hit(
            self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy
    ) -> Dict[str, Decimal]:
        price = None
        amount = min(self.amount_pending, liquidity_strategy.available_liquidity)
        base_sign = helpers.get_base_sign_for_operation(self.operation)

        if self.operation == OrderOperation.BUY:
            # Limit price was hit at bar open.
            if bar.open < self._limit_price:
                price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_high=self._limit_price)
            # The price went down to limit price or lower.
            elif bar.low <= self._limit_price:
                price = self._limit_price
        else:
            assert self.operation == OrderOperation.SELL
            # Limit price was hit at bar open.
            if bar.open > self._limit_price:
                price = slipped_price(bar.open, self.operation, amount, liquidity_strategy, cap_low=self._limit_price)
            # The price went up to limit price or higher.
            elif bar.high >= self._limit_price:
                price = self._limit_price

        ret = {}
        if amount and price:
            ret = {
                self.pair.base_symbol: amount * base_sign,
                self.pair.quote_symbol: price * amount * -base_sign
            }
        return ret

    def get_balance_updates(
            self, bar: bar.Bar, liquidity_strategy: liquidity.LiquidityStrategy
    ) -> Dict[str, Decimal]:
        if not self._stop_price_hit:
            ret = self.get_balance_updates_before_stop_hit(bar, liquidity_strategy)
        else:
            ret = self.get_balance_updates_after_stop_hit(bar, liquidity_strategy)
        return ret

    def calculate_estimated_fill_price(self) -> Optional[Decimal]:
        # It will be the limit price or a better one.
        return self._limit_price

    def get_debug_info(self) -> dict:
        ret = super().get_debug_info()
        ret["limit_price"] = self._limit_price
        ret["stop_price"] = self._stop_price
        return ret

    def get_order_info(self) -> OrderInfo:
        ret = super().get_order_info()
        ret.limit_price = self._limit_price
        ret.stop_price = self._stop_price
        return ret


def slipped_price(
        price: Decimal, operation: OrderOperation, amount: Decimal, liquidity_strategy: liquidity.LiquidityStrategy,
        cap_low: Optional[Decimal] = None, cap_high: Optional[Decimal] = None
) -> Decimal:
    price_impact = liquidity_strategy.calculate_price_impact(amount)
    if operation == OrderOperation.BUY:
        price *= (Decimal(1) + price_impact)
    else:
        assert operation == OrderOperation.SELL
        price *= (Decimal(1) - price_impact)

    if cap_low is not None:
        price = max(price, cap_low)
    if cap_high is not None:
        price = min(price, cap_high)

    return price



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

from .core.bar import (
    Bar,
    BarEvent,
)

from .core.dispatcher import (
    EventDispatcher,
    BacktestingDispatcher,
    RealtimeDispatcher,
    backtesting_dispatcher,
    realtime_dispatcher,
)

from .core.dt import (
    local_now,
    utc_now,
)

from .core.event import (
    Event,
    EventSource,
    FifoQueueEventSource,
    Producer,
)

from .core.event_sources.trading_signal import (
    TradingSignal,
    TradingSignalSource,
)

from .core.enums import (
    OrderOperation,
    Position,
)

from .core.helpers import (
    round_decimal,
    truncate_decimal,
)

from .core.pair import (
    Pair,
    PairInfo,
)

from .core.token_bucket import (
    TokenBucketLimiter,
)




# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, List, Optional, Tuple, Union
import dataclasses

import aiohttp

from . import client, helpers, order_book, trades, spot, cross_margin, isolated_margin, websocket_mgr
from basana.core import bar, dispatcher, enums, token_bucket
from basana.core.pair import Pair, PairInfo


BarEventHandler = bar.BarEventHandler
Error = client.Error
OrderBookEvent = order_book.OrderBookEvent
OrderBookEventHandler = order_book.OrderBookEventHandler
OrderOperation = enums.OrderOperation
TradeEvent = trades.TradeEvent
TradeEventHandler = trades.TradeEventHandler


@dataclasses.dataclass(frozen=True)
class PairInfoEx(PairInfo):
    """Information about a trading pair.

    :param base_precision: The precision for the base symbol.
    :param quote_precision: The precision for the quote symbol.
    :param permissions: The account and pair permissions.

    Check **Account and Symbol Permissions** in https://binance-docs.github.io/apidocs/spot/en/#public-api-definitions.
    """

    #: The account and pair permissions.
    permissions: List[str]


class Exchange:
    """A client for `Binance <https://www.binance.com/>`_ crypto currency exchange.

    :param dispatcher: The event dispatcher.
    :param api_key: An optional api key. If not set only public endpoints can be used.
    :param api_secret: An optional api secret. If not set only public endpoints can be used.
    :param session: An optional client session, in case you want to reuse connections.
    :type session: aiohttp.ClientSession
    :param tb: An optional token bucket limiter, in case you want to throttle requests.
    :param config_overrides: An optional dictionary for overriding config settings.
    """

    def __init__(
            self, dispatcher: dispatcher.EventDispatcher, api_key: Optional[str] = None,
            api_secret: Optional[str] = None, session: Optional[aiohttp.ClientSession] = None,
            tb: Optional[token_bucket.TokenBucketLimiter] = None, config_overrides: dict = {}
    ):
        self._dispatcher = dispatcher
        self._cli = client.APIClient(
            api_key=api_key, api_secret=api_secret, session=session, tb=tb, config_overrides=config_overrides
        )
        self._session = session
        self._tb = tb
        self._config_overrides = config_overrides
        self._pair_info_cache: Dict[Pair, PairInfoEx] = {}
        self._ws_mgr = websocket_mgr.WebsocketManager(
            dispatcher, self._cli, session=session, config_overrides=config_overrides
        )

    def subscribe_to_bar_events(
            self, pair: Pair, bar_duration: Union[int, str], event_handler: BarEventHandler,
            skip_first_bar: bool = True, flush_delay: float = 1
    ):
        """
        Registers an async callable that will be called when a new bar is available.

        Works as defined in https://binance-docs.github.io/apidocs/spot/en/#kline-candlestick-streams but only closed
        klines are reported.

        :param pair: The trading pair.
        :param bar_duration: The bar duration. One of 1s, 1m, 3m, 5m, 15m, 30m, 1h, 2h, 4h, 6h, 8h, 12h, 1d, 3d, 1w, 1M.
        :type bar_duration: str
        :param event_handler: An async callable that receives a BarEvent.
        :param skip_first_bar: Ignored. It will be removed in a future version.
        :param flush_delay: Ignored. It will be removed in a future version.
        """
        # TODO: Deprecate support for bar_duration as int.
        # TODO: Remove skip_first_bar and flush_delay.
        interval = {
            # Supporting interval as int for backwards compatibility reasons.
            1: "1s",
            60: "1m",
            3 * 60: "3m",
            5 * 60: "5m",
            15 * 60: "15m",
            30 * 60: "30m",
            3600: "1h",
            2 * 3600: "2h",
            4 * 3600: "4h",
            6 * 3600: "6h",
            8 * 3600: "8h",
            12 * 3600: "12h",
            86400: "1d",
            3 * 86400: "3d",
            7 * 86400: "1w",
            31 * 86400: "1M",
            # Once support for interval as int is removed, this should be simplified.
            "1s": "1s",
            "1m": "1m",
            "3m": "3m",
            "5m": "5m",
            "15m": "15m",
            "30m": "30m",
            "1h": "1h",
            "2h": "2h",
            "4h": "4h",
            "6h": "6h",
            "8h": "8h",
            "12h": "12h",
            "1d": "1d",
            "3d": "3d",
            "1w": "1w",
            "1M": "1M",
        }.get(bar_duration)
        assert interval, "Invalid bar_duration"

        self._ws_mgr.subscribe_to_bar_events(pair, interval, event_handler)

    def subscribe_to_order_book_events(
            self, pair: Pair, event_handler: OrderBookEventHandler, depth: int = 10
    ):
        """
        Registers an async callable that will be called every 1 second with the top bids/asks of the order book.

        Works as defined in https://binance-docs.github.io/apidocs/spot/en/#partial-book-depth-streams.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives an OrderBookEvent.
        :param depth: The order book depth. Valid values are: 5, 10, 20.
        """

        self._ws_mgr.subscribe_to_order_book_events(pair, event_handler, depth=depth)

    def subscribe_to_trade_events(self, pair: Pair, event_handler: TradeEventHandler):
        """
        Registers an async callable that will be called for every new trade.

        Works as defined in https://binance-docs.github.io/apidocs/spot/en/#trade-streams.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives a TradeEvent.
        """

        self._ws_mgr.subscribe_to_trade_events(pair, event_handler)

    async def get_pair_info(self, pair: Pair) -> PairInfoEx:
        """Returns information about a trading pair.

        :param pair: The trading pair.
        """
        ret = self._pair_info_cache.get(pair)
        if not ret:
            exchange_info = await self._cli.get_exchange_info(helpers.pair_to_order_book_symbol(pair))
            symbols = exchange_info["symbols"]
            assert len(symbols) == 1, "More than 1 symbol found"
            symbol_info = symbols[0]
            price_filter = get_filter_from_symbol_info(symbol_info, "PRICE_FILTER")
            assert price_filter, f"PRICE_FILTER not found for {pair}"
            lot_size = get_filter_from_symbol_info(symbol_info, "LOT_SIZE")
            assert lot_size, f"LOT_SIZE not found for {pair}"
            ret = PairInfoEx(
                base_precision=get_precision_from_step_size(lot_size["stepSize"]),
                quote_precision=get_precision_from_step_size(price_filter["tickSize"]),
                permissions=symbol_info.get("permissions")
            )
            self._pair_info_cache[pair] = ret
        return ret

    async def get_bid_ask(self, pair: Pair) -> Tuple[Decimal, Decimal]:
        """Returns the current bid and ask price.

        :param pair: The trading pair.
        """
        order_book = await self._cli.get_order_book(helpers.pair_to_order_book_symbol(pair), limit=1)
        return Decimal(order_book["bids"][0][0]), Decimal(order_book["asks"][0][0])

    @property
    def spot_account(self) -> spot.Account:
        """Returns the spot account."""
        return spot.Account(self._cli.spot_account, self._ws_mgr)

    @property
    def cross_margin_account(self) -> cross_margin.Account:
        """Returns the cross margin account."""
        return cross_margin.Account(self._cli.cross_margin_account, self._ws_mgr)

    @property
    def isolated_margin_account(self) -> isolated_margin.Account:
        """Returns the isolated margin account."""
        return isolated_margin.Account(self._cli.isolated_margin_account, self._ws_mgr)


def get_filter_from_symbol_info(symbol_info: dict, filter_type: str) -> Optional[dict]:
    filters = symbol_info["filters"]
    price_filters = [filter for filter in filters if filter["filterType"] == filter_type]
    return None if not price_filters else price_filters[0]


def get_precision_from_step_size(step_size: str) -> int:
    return int(-Decimal(step_size).log10() / Decimal(10).log10())



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, Optional
import abc

from . import helpers
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


class ExchangeOrder(metaclass=abc.ABCMeta):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Optional[Decimal],
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ):
        self._operation = operation
        self._pair = pair
        self._amount = amount
        self._client_order_id = client_order_id
        self._kwargs = kwargs

    @abc.abstractmethod
    async def create_order(self, margin_account_cli) -> dict:
        raise NotImplementedError()


class MarketOrder(ExchangeOrder):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Optional[Decimal] = None,
            quote_amount: Optional[Decimal] = None, client_order_id: Optional[str] = None,
            side_effect_type: str = "NO_SIDE_EFFECT", **kwargs: Dict[str, Any]
    ):
        assert (amount is not None) ^ (quote_amount is not None), "Either amount or quote_amount should be set"
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)
        self._quote_amount = quote_amount
        self._side_effect_type = side_effect_type

    async def create_order(self, margin_account_cli) -> dict:
        return await margin_account_cli.create_order(
            helpers.pair_to_order_book_symbol(self._pair), helpers.order_operation_to_side(self._operation), "MARKET",
            quantity=self._amount, quote_order_qty=self._quote_amount, new_client_order_id=self._client_order_id,
            side_effect_type=self._side_effect_type, **self._kwargs
        )


class LimitOrder(ExchangeOrder):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            side_effect_type: str = "NO_SIDE_EFFECT", time_in_force: str = "GTC", client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ):
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)
        self._limit_price = limit_price
        self._time_in_force = time_in_force
        self._side_effect_type = side_effect_type

    async def create_order(self, margin_account_cli) -> dict:
        return await margin_account_cli.create_order(
            helpers.pair_to_order_book_symbol(self._pair), helpers.order_operation_to_side(self._operation), "LIMIT",
            quantity=self._amount, price=self._limit_price, time_in_force=self._time_in_force,
            new_client_order_id=self._client_order_id, side_effect_type=self._side_effect_type, **self._kwargs
        )


class StopLimitOrder(ExchangeOrder):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal, limit_price: Decimal,
            side_effect_type: str = "NO_SIDE_EFFECT", time_in_force: str = "GTC",
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ):
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)
        self._stop_price = stop_price
        self._limit_price = limit_price
        self._time_in_force = time_in_force
        self._side_effect_type = side_effect_type

    async def create_order(self, margin_account_cli) -> dict:
        return await margin_account_cli.create_order(
            helpers.pair_to_order_book_symbol(self._pair), helpers.order_operation_to_side(self._operation),
            "STOP_LOSS_LIMIT", quantity=self._amount, stop_price=self._stop_price, price=self._limit_price,
            time_in_force=self._time_in_force, new_client_order_id=self._client_order_id,
            side_effect_type=self._side_effect_type, **self._kwargs
        )


class OCOOrder:
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal, stop_price: Decimal,
            stop_limit_price: Optional[Decimal] = None, side_effect_type: str = "NO_SIDE_EFFECT",
            stop_limit_time_in_force: str = "GTC", list_client_order_id: Optional[str] = None,
            limit_client_order_id: Optional[str] = None, stop_client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ):
        self._operation = operation
        self._pair = pair
        self._amount = amount
        self._limit_price = limit_price
        self._stop_price = stop_price
        self._stop_limit_price = stop_limit_price
        self._stop_limit_time_in_force = None if stop_limit_price is None else stop_limit_time_in_force
        self._list_client_order_id = list_client_order_id
        self._side_effect_type = side_effect_type
        self._limit_client_order_id = limit_client_order_id
        self._stop_client_order_id = stop_client_order_id
        self._kwargs = kwargs

    async def create_order(self, margin_account_cli) -> dict:
        return await margin_account_cli.create_oco(
            helpers.pair_to_order_book_symbol(self._pair), helpers.order_operation_to_side(self._operation),
            self._amount, self._limit_price, self._stop_price, stop_limit_price=self._stop_limit_price,
            stop_limit_time_in_force=self._stop_limit_time_in_force, list_client_order_id=self._list_client_order_id,
            side_effect_type=self._side_effect_type, limit_client_order_id=self._limit_client_order_id,
            stop_client_order_id=self._stop_client_order_id, **self._kwargs
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Missing:
    pass


def _get_config_value_impl(config: dict, path: str, default=None):
    ret = default
    current_dict = config
    keys = path.split(".")
    for i, key in enumerate(keys):
        assert key, "Invalid path {}".format(path)
        if i == len(keys) - 1:
            ret = current_dict.get(key, default)
        else:
            current_dict = current_dict.get(key, {})
            assert isinstance(current_dict, dict), f"Element at {key} is not a dictionary"
    return ret


def get_config_value(config: dict, path: str, default=None, overrides: dict = {}):
    missing = Missing()
    ret = _get_config_value_impl(overrides, path, default=missing)
    if ret == missing:
        ret = _get_config_value_impl(config, path, default=default)
    return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
import abc

from basana.backtesting import errors, orders
from basana.core import helpers
from basana.core.enums import OrderOperation
from basana.core.pair import Pair, PairInfo


class ExchangeOrder(metaclass=abc.ABCMeta):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        self._operation = operation
        self._pair = pair
        self._amount = amount
        self._auto_borrow = auto_borrow
        self._auto_repay = auto_repay

    @property
    def pair(self) -> Pair:
        return self._pair

    @property
    def amount(self) -> Decimal:
        return self._amount

    @property
    def operation(self) -> OrderOperation:
        return self._operation

    @property
    def auto_borrow(self) -> bool:
        return self._auto_borrow

    @property
    def auto_repay(self) -> bool:
        return self._auto_repay

    def validate(self, pair_info: PairInfo):
        if self.amount <= Decimal(0):
            raise errors.Error("Amount must be > 0")
        if self.amount != helpers.truncate_decimal(self.amount, pair_info.base_precision):
            raise errors.Error(
                "{} exceeds maximum precision of {} decimal digits".format(self.amount, pair_info.base_precision)
            )

    @abc.abstractmethod
    def create_order(self, id: str) -> orders.Order:
        raise NotImplementedError()


class MarketOrder(ExchangeOrder):
    """Market order request.

    A market order is an order to buy or sell a stock at the best available price.
    Generally, this type of order will be executed immediately. However, the price at which a market order will be
    executed is not guaranteed.
    """

    def validate(self, pair_info: PairInfo):
        super().validate(pair_info)

    def create_order(self, id: str) -> orders.Order:
        return orders.MarketOrder(
            id, self.operation, self.pair, self.amount, orders.OrderState.OPEN, auto_borrow=self.auto_borrow,
            auto_repay=self.auto_repay
        )


class LimitOrder(ExchangeOrder):
    """Limit order request.

    A limit order is an order to buy or sell a stock at a specific price or better.
    A buy limit order can only be executed at the limit price or lower, and a sell limit order can only be executed
    at the limit price or higher.
    """

    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        super().__init__(operation, pair, amount, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._limit_price = limit_price

    @property
    def limit_price(self) -> Decimal:
        return self._limit_price

    def validate(self, pair_info: PairInfo):
        super().validate(pair_info)
        if self.limit_price <= Decimal(0):
            raise errors.Error("Limit price must be > 0")
        if self.limit_price != helpers.truncate_decimal(self.limit_price, pair_info.quote_precision):
            raise errors.Error(
                "{} exceeds maximum precision of {} decimal digits".format(self.limit_price, pair_info.quote_precision)
            )

    def create_order(self, id: str) -> orders.Order:
        return orders.LimitOrder(
            id, self.operation, self.pair, self.amount, self._limit_price, orders.OrderState.OPEN,
            auto_borrow=self.auto_borrow, auto_repay=self.auto_repay
        )


class StopOrder(ExchangeOrder):
    """Stop order request.

    A stop order, also referred to as a stop-loss order, is an order to buy or sell a stock once the price of the
    stock reaches a specified price, known as the stop price.
    When the stop price is reached, a stop order becomes a market order.
    A buy stop order is entered at a stop price above the current market price. Investors generally use a buy stop
    order to limit a loss or to protect a profit on a stock that they have sold short.
    A sell stop order is entered at a stop price below the current market price. Investors generally use a sell
    stop order to limit a loss or to protect a profit on a stock that they own.
    """

    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        super().__init__(operation, pair, amount, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._stop_price = stop_price

    @property
    def stop_price(self) -> Decimal:
        return self._stop_price

    def validate(self, pair_info: PairInfo):
        super().validate(pair_info)
        if self.stop_price <= Decimal(0):
            raise errors.Error("Stop price must be > 0")
        if self.stop_price != helpers.truncate_decimal(self.stop_price, pair_info.quote_precision):
            raise errors.Error(
                "{} exceeds maximum precision of {} decimal digits".format(self.stop_price, pair_info.quote_precision)
            )

    def create_order(self, id: str) -> orders.Order:
        return orders.StopOrder(
            id, self.operation, self.pair, self.amount, self._stop_price, orders.OrderState.OPEN,
            auto_borrow=self.auto_borrow, auto_repay=self.auto_repay
        )


class StopLimitOrder(ExchangeOrder):
    """Stop limit order request.

    A stop-limit order is an order to buy or sell a stock that combines the features of a stop order and a limit
    order.
    Once the stop price is reached, a stop-limit order becomes a limit order that will be executed at a specified
    price (or better). The benefit of a stop-limit order is that the investor can control the price at which the
    order can be executed.
    """

    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal, limit_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ):
        super().__init__(operation, pair, amount, auto_borrow=auto_borrow, auto_repay=auto_repay)
        self._stop_price = stop_price
        self._limit_price = limit_price

    @property
    def stop_price(self) -> Decimal:
        return self._stop_price

    @property
    def limit_price(self) -> Decimal:
        return self._limit_price

    def validate(self, pair_info: PairInfo):
        super().validate(pair_info)
        if self.stop_price <= Decimal(0):
            raise errors.Error("Stop price must be > 0")
        if self.limit_price <= Decimal(0):
            raise errors.Error("Limit price must be > 0")
        for value in [self.stop_price, self.limit_price]:
            if value != helpers.truncate_decimal(value, pair_info.quote_precision):
                raise errors.Error(
                    "{} exceeds maximum precision of {} decimal digits".format(value, pair_info.quote_precision)
                )

    def create_order(self, id: str) -> orders.Order:
        return orders.StopLimitOrder(
            id, self.operation, self.pair, self.amount, self._stop_price, self._limit_price, orders.OrderState.OPEN,
            auto_borrow=self.auto_borrow, auto_repay=self.auto_repay
        )
